; See also perm-transitive.lisp for the result of
; cleaning up this demo to create a certifiable
; book.

(defun mem (a lst)
; Return t if a is a member of the list, lst, else
; nil.  This definition is typical in style for
; list processing.  The termination proof
; obligation, important for soundness, is
; discharged automatically.
  (cond ((atom lst)
         nil)
        ((equal a (car lst))
         t)
        (t
         (mem a (cdr lst)))))

(defun del (a lst)
; This function deletes the first occurrence of a
; (if any) in the list, lst.
  (cond ((atom lst)
         nil)
        ((equal a (car lst))
         (cdr lst))
        (t
         (cons (car lst)
               (del a (cdr lst))))))

(defun perm (lst1 lst2)
; Given two lists, return t if lst1 is a
; permutation of lst2.
  (cond ((atom lst1)
         (atom lst2))
        ((mem (car lst1) lst2)
         (perm (cdr lst1)
               (del (car lst1) lst2)))
        (t
         nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Prove that perm is transitive.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; fails
(defthm perm-transitive ; OUR GOAL for this demo
  (implies (and (perm x y) (perm y z))
           (perm x z)))

#|| first key checkpoint under induction for
    attempt to prove perm-transitive:
 Subgoal *1/5''
 (IMPLIES (AND (CONSP X)
               (NOT (MEM (CAR X) Z))
               (MEM (CAR X) Y)
               (PERM (CDR X) (DEL (CAR X) Y)))
          (NOT (PERM Y Z)))
||#

; fails
(defthm not-perm-if-different-members
  ;; lemma for perm-transitive
  (implies (and (not (mem a z))
                (mem a y))
           (equal (perm y z) nil)))

#|| first key checkpoint under induction for
    attempt to prove not-perm-if-different-members:
 Subgoal *1/2''
 (IMPLIES (AND (CONSP Y)
               (MEM (CAR Y) Z)
               (MEM A (DEL (CAR Y) Z))
               (NOT (MEM A Z))
               (MEM A (CDR Y)))
          (NOT (PERM (CDR Y) (DEL (CAR Y) Z))))
||#

; succeeds
; Contraposed form avoids free variables in the hypothesis.
(defthm mem-del-implies-mem
  ;; lemma for not-perm-if-different-members
  (implies (not (mem a z))
           (equal (mem a (del b z))
                  nil)))

; now succeeds
(defthm not-perm-if-different-members
  ;; lemma for perm-transitive
  (implies (and (not (mem a z))
                (mem a y))
           (equal (perm y z) nil)))

; still fails
(defthm perm-transitive ; OUR GOAL for this demo
  (implies (and (perm x y) (perm y z))
           (perm x z)))

#|| first key checkpoint under induction for
    attempt to prove perm-transitive:
 Subgoal *1/3''
 (IMPLIES (AND (CONSP X)
               (MEM (CAR X) Z)
               (NOT (PERM (DEL (CAR X) Y) (DEL (CAR X) Z)))
               (MEM (CAR X) Y)
               (PERM (CDR X) (DEL (CAR X) Y))
               (PERM Y Z))
          (PERM (CDR X) (DEL (CAR X) Z)))
||#

; fails
(defthm perm-del-del ; lemma for perm-transitive
  (implies (perm y z)
           (perm (del a y) (del a z))))

#|| first key checkpoint under induction for
    attempt to prove perm-del-del:
 Subgoal *1/4.2
 (IMPLIES (AND (CONSP Y)
               (NOT (EQUAL A (CAR Y)))
               (PERM (DEL A (CDR Y))
                     (DEL A (DEL (CAR Y) Z)))
               (MEM (CAR Y) Z)
               (PERM (CDR Y) (DEL (CAR Y) Z)))
          (MEM (CAR Y) (DEL A Z)))
||#

; succeeds
(defthm mem-del
  ;; lemma for perm-del-del
  (implies (not (equal a b))
           (equal (mem b (del a z))
                  (mem b z))))

; still fails
(defthm perm-del-del
  ;; lemma for perm-transitive
  (implies (perm y z)
           (perm (del a y) (del a z))))

#|| first key checkpoint under induction for perm-del-del:
 Subgoal *1/4''
 (IMPLIES (AND (CONSP Y)
               (NOT (EQUAL A (CAR Y)))
               (PERM (DEL A (CDR Y))
                     (DEL A (DEL (CAR Y) Z)))
               (MEM (CAR Y) Z)
               (PERM (CDR Y) (DEL (CAR Y) Z)))
          (PERM (DEL A (CDR Y))
                (DEL (CAR Y) (DEL A Z))))
||#

(defthm del-del
  ;; lemma for perm-del-del (NOTE: DOESN'T LOOP!)
  (equal (del a (del b x))
         (del b (del a x))))

; now succeeds
(defthm perm-del-del
  ;; lemma for perm-transitive
  (implies (perm y z)
           (perm (del a y) (del a z))))

(defthm perm-transitive
  (implies (and (perm x y) (perm y z))
           (perm x z)))
