; See perm-transitive-demo.lsp for a demo, with comments explaining the proof
; development process.  The present file is the result of cleaning up that demo
; file to create a certifiable book.

(in-package "ACL2")

(defun mem (a lst)
; Return t if a is a member of the list, lst, else
; nil.  This definition is typical in style for
; list processing.  The termination proof
; obligation, important for soundness, is
; discharged automatically.
  (cond ((atom lst)
         nil)
        ((equal a (car lst))
         t)
        (t
         (mem a (cdr lst)))))

(defun del (a lst)
; This function deletes the first occurrence of a
; (if any) in the list, lst.
  (cond ((atom lst)
         nil)
        ((equal a (car lst))
         (cdr lst))
        (t
         (cons (car lst)
               (del a (cdr lst))))))

(defun perm (lst1 lst2)
; Given two lists, return t if lst1 is a
; permutation of lst2.
  (cond ((atom lst1)
         (atom lst2))
        ((mem (car lst1) lst2)
         (perm (cdr lst1)
               (del (car lst1) lst2)))
        (t
         nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Prove that the permutation relation is transitive.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defthm mem-del-implies-mem ; for not-perm-if-different-members
; Contraposed form avoids free variables in the hypothesis.
  (implies (not (mem a z))
           (not (mem a (del b z)))))

(defthm not-perm-if-different-members ; lemma for perm-transitive
  (implies (and (not (mem a z))
                (mem a y))
           (not (perm y z))))

(defthm mem-del ; for perm-del-del
  (implies (and (not (equal a b))
                (mem b z))
           (mem b (del a z))))

(defthm del-del ; for perm-del-del (NOTE: DOESN'T LOOP!)
  (equal (del a (del b x))
         (del b (del a x))))

(defthm perm-del-del ; lemma for perm-transitive
  (implies (perm y z)
           (perm (del a y) (del a z))))

(defthm perm-transitive
  (implies (and (perm x y) (perm y z))
           (perm x z)))
